var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _AbstractGossipLog_apply;
import { TypedEventEmitter, CustomEvent, CodeError } from "@libp2p/interface";
import { logger } from "@libp2p/logger";
import { equals, toString } from "uint8arrays";
import { hashEntry } from "@canvas-js/okra";
import { ed25519 } from "@canvas-js/signatures";
import { assert, zip, prepare, prepareMessage } from "@canvas-js/utils";
import { Driver } from "./sync/driver.js";
import { AncestorIndex } from "./AncestorIndex.js";
import { BranchMergeIndex } from "./BranchMergeIndex.js";
import { SignedMessage } from "./SignedMessage.js";
import { decodeId, encodeId, messageIdPattern } from "./ids.js";
import { getNextClock } from "./schema.js";
import { MISSING_PARENT, topicPattern } from "./utils.js";
import { GossipLogService } from "./GossipLogService.js";
export class AbstractGossipLog extends TypedEventEmitter {
    constructor(init) {
        super();
        this.libp2p = null;
        this.service = null;
        _AbstractGossipLog_apply.set(this, void 0);
        this.handlePruneEvent = ({ detail: peer }) => {
            if (peer.topic === this.topic) {
                this.dispatchEvent(new CustomEvent("prune", { detail: { peerId: peer.peerId.toString() } }));
            }
        };
        this.handleGraftEvent = ({ detail: peer }) => {
            if (peer.topic === this.topic) {
                this.dispatchEvent(new CustomEvent("graft", { detail: { peerId: peer.peerId.toString() } }));
            }
        };
        assert(topicPattern.test(init.topic), "invalid topic (must match [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        this.signer = init.signer ?? ed25519.create();
        __classPrivateFieldSet(this, _AbstractGossipLog_apply, init.apply, "f");
        this.validatePayload = init.validatePayload ?? ((payload) => true);
        this.verifySignature = init.verifySignature ?? this.signer.scheme.verify;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async listen(libp2p) {
        assert(libp2p.status === "started");
        assert(this.libp2p === null);
        assert(this.service === null);
        this.libp2p = libp2p;
        this.service = new GossipLogService(libp2p, this, {});
        this.service.pubsub?.addEventListener("gossipsub:prune", this.handlePruneEvent);
        this.service.pubsub?.addEventListener("gossipsub:graft", this.handleGraftEvent);
        await this.service.start();
    }
    async replay() {
        await this.tree.read(async (txn) => {
            for (const leaf of txn.keys()) {
                const id = decodeId(leaf);
                const record = await this.db.get("$messages", id);
                if (record === null) {
                    this.log.error("failed to get message %s from database", id);
                    continue;
                }
                const signedMessage = this.encode(record.signature, record.message);
                assert(signedMessage.id === id);
                await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(this, [signedMessage, record.branch]);
            }
        });
    }
    encode(signature, message) {
        assert(this.topic === message.topic, "expected this.topic === topic");
        const preparedMessage = prepareMessage(message);
        assert(this.validatePayload(preparedMessage.payload), "error encoding message (invalid payload)");
        return SignedMessage.encode(signature, preparedMessage);
    }
    decode(value) {
        const signedMessage = SignedMessage.decode(value);
        assert(this.topic === signedMessage.message.topic, "expected this.topic === topic");
        assert(this.validatePayload(signedMessage.message.payload), "error decoding message (invalid payload)");
        return signedMessage;
    }
    async getClock() {
        const heads = await this.db.query("$heads", { orderBy: { id: "asc" } });
        const ids = heads.map(({ id }) => id);
        const clock = getNextClock(ids.map(encodeId));
        return [clock, ids];
    }
    async has(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const records = await this.db.query("$messages", { select: { id: true }, where: { id } });
        return records.length > 0;
    }
    async get(id) {
        const record = await this.db.get("$messages", id);
        if (record === null) {
            return [null, null];
        }
        else {
            return [record.signature, record.message];
        }
    }
    getMessages(range = {}) {
        const { reverse = false, limit, ...where } = range;
        return this.db.query("$messages", {
            where: { id: where },
            select: { id: true, signature: true, message: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        });
    }
    async *iterate(range = {}) {
        const { reverse = false, limit, ...where } = range;
        // TODO: use this.db.iterate()
        const query = await this.db.query("$messages", {
            where: { id: where },
            select: { id: true, signature: true, message: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        });
        for await (const row of query) {
            yield row;
        }
    }
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The concurrent heads of the local log are used as parents.
     */
    async append(payload, { signer = this.signer, publish = true } = {}) {
        let root = null;
        let heads = null;
        const signedMessage = await this.tree.write(async (txn) => {
            const [clock, parents] = await this.getClock();
            const message = {
                topic: this.topic,
                clock,
                parents,
                payload: prepare(payload, { replaceUndefined: true }),
            };
            const signature = await signer.sign(message);
            const signedMessage = this.encode(signature, message);
            this.log("appending message %s at clock %d with parents %o", signedMessage.id, clock, parents);
            const result = await this.apply(txn, signedMessage);
            root = result.root;
            heads = result.heads;
            return signedMessage;
        });
        assert(root !== null && heads !== null, "failed to commit transaction");
        this.dispatchEvent(new CustomEvent("commit", { detail: { root, heads } }));
        if (publish && this.service !== null) {
            return {
                id: signedMessage.id,
                signature: signedMessage.signature,
                message: signedMessage.message,
                recipients: this.service.publish(signedMessage),
            };
        }
        else {
            return {
                id: signedMessage.id,
                signature: signedMessage.signature,
                message: signedMessage.message,
                recipients: Promise.resolve([]),
            };
        }
    }
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    async insert(signedMessage, { publish = true, ...options } = {}) {
        const { message, signature } = signedMessage;
        assert(message.topic === this.topic, `expected message.topic === this.topic`);
        await this.verifySignature(signature, message);
        const { clock, parents } = signedMessage.message;
        const signedMessageInstance = signedMessage instanceof SignedMessage ? signedMessage : this.encode(signature, message);
        const id = signedMessageInstance.id;
        this.log("inserting message %s at clock %d with parents %o", id, clock, parents);
        const parentKeys = message.parents.map(encodeId);
        let root = null;
        let heads = null;
        await this.tree.write(async (txn) => {
            const hasSignedMessage = await this.has(id);
            if (hasSignedMessage) {
                return;
            }
            for (const parentKey of parentKeys) {
                const leaf = txn.getNode(0, parentKey);
                if (leaf === null) {
                    this.log.error("missing parent %s of message %s: %O", parent, id, message);
                    throw new CodeError(`missing parent ${parent} of message ${id}`, MISSING_PARENT);
                }
            }
            const result = await this.apply(txn, signedMessageInstance);
            root = result.root;
            heads = result.heads;
        });
        assert(root !== null && heads !== null, "failed to commit transaction");
        this.dispatchEvent(new CustomEvent("commit", { detail: { root, heads } }));
        if (publish && this.service !== null) {
            const recipients = this.service.publish(signedMessageInstance);
            return { id, recipients };
        }
        else {
            return { id, recipients: Promise.resolve([]) };
        }
    }
    async apply(txn, signedMessage) {
        const { id, signature, message, key, value } = signedMessage;
        this.log("applying %s %O", id, message);
        const parentMessageRecords = [];
        for (const parentId of message.parents) {
            const parentMessageRecord = await this.db.get("$messages", parentId);
            if (parentMessageRecord === null) {
                throw new Error(`missing parent ${parentId} of message ${id}`);
            }
            parentMessageRecords.push(parentMessageRecord);
        }
        const branch = await this.getBranch(id, parentMessageRecords);
        try {
            await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(this, [signedMessage, branch]);
        }
        catch (error) {
            this.dispatchEvent(new CustomEvent("error", { detail: { error } }));
            throw error;
        }
        const hash = toString(hashEntry(key, value), "hex");
        const branchMergeIndex = new BranchMergeIndex(this.db);
        for (const parentMessageRecord of parentMessageRecords) {
            if (parentMessageRecord.branch !== branch) {
                await branchMergeIndex.insertBranchMerge({
                    source_branch: parentMessageRecord.branch,
                    source_clock: parentMessageRecord.clock,
                    source_message_id: parentMessageRecord.id,
                    target_branch: branch,
                    target_clock: message.clock,
                    target_message_id: id,
                });
            }
        }
        const messageRecord = { id, signature, message, hash, branch, clock: message.clock };
        const heads = await this.db
            .query("$heads", { select: { id: true } })
            .then((heads) => heads.map((head) => head.id))
            .then((heads) => heads.filter((head) => message.parents.includes(head)));
        await this.db.apply([
            ...heads.map((head) => ({ model: "$heads", operation: "delete", key: head })),
            { model: "$heads", operation: "set", value: { id } },
            { model: "$messages", operation: "set", value: messageRecord },
        ]);
        txn.set(key, value);
        await new AncestorIndex(this.db).indexAncestors(id, message.parents);
        this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message } }));
        const root = txn.getRoot();
        return { root, heads };
    }
    async newBranch() {
        const maxBranchRecords = await this.db.query("$messages", {
            select: { id: true, branch: true },
            limit: 1,
            orderBy: { branch: "desc" },
        });
        if (maxBranchRecords.length == 0) {
            return 0;
        }
        else {
            return maxBranchRecords[0].branch + 1;
        }
    }
    async getBranch(messageId, parentMessages) {
        if (parentMessages.length == 0) {
            return await this.newBranch();
        }
        const parentMessageWithMaxClock = parentMessages.reduce((max, parentMessage) => max.branch > parentMessage.branch ? max : parentMessage);
        const branch = parentMessageWithMaxClock.branch;
        const messagesAtBranchClockPosition = await this.db.query("$messages", {
            select: { id: true, branch: true, clock: true },
            where: {
                branch,
                clock: { gt: parentMessageWithMaxClock.clock },
                id: { neq: messageId },
            },
            limit: 1,
        });
        if (messagesAtBranchClockPosition.length > 0) {
            return await this.newBranch();
        }
        else {
            return branch;
        }
    }
    async isAncestor(id, ancestor, visited = new Set()) {
        return await new AncestorIndex(this.db).isAncestor(id, ancestor, visited);
    }
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    async sync(server, callback = async (signedMessage) => {
        await this.insert(signedMessage);
    }, options = {}) {
        const start = performance.now();
        let messageCount = 0;
        await this.tree.read(async (txn) => {
            const driver = new Driver(this.topic, server, txn);
            for await (const keys of driver.sync()) {
                const values = await server.getValues(keys);
                for (const [key, value] of zip(keys, values)) {
                    const signedMessage = this.decode(value);
                    assert(equals(key, signedMessage.key), "invalid message key");
                    await callback(signedMessage);
                    messageCount++;
                }
            }
        });
        const duration = Math.ceil(performance.now() - start);
        this.log("finished sync with peer %s (%d messages in %dms)", options.peerId, messageCount, duration);
        this.dispatchEvent(new CustomEvent("sync", { detail: { peerId: options.peerId, messageCount, duration } }));
        return { messageCount };
    }
    serve(callback) {
        return this.tree.read((txn) => callback({
            getRoot: () => txn.getRoot(),
            getNode: (level, key) => txn.getNode(level, key),
            getChildren: (level, key) => txn.getChildren(level, key),
            getValues: async (keys) => {
                const values = [];
                // TODO: txn.getMany
                for (const key of keys) {
                    const id = decodeId(key);
                    const [signature, message] = await this.get(id);
                    if (signature === null || message === null) {
                        throw new CodeError("message not found", "NOT_FOUND", { id });
                    }
                    const signedMessage = SignedMessage.encode(signature, message);
                    assert(equals(signedMessage.key, key), "invalid message key");
                    values.push(signedMessage.value);
                }
                return values;
            },
        }));
    }
}
_AbstractGossipLog_apply = new WeakMap();
AbstractGossipLog.schema = {
    $messages: {
        id: "primary",
        signature: "json",
        message: "json",
        hash: "string",
        branch: "integer",
        clock: "integer",
        $indexes: ["branch", "clock"],
    },
    $heads: { id: "primary" },
    ...AncestorIndex.schema,
    ...BranchMergeIndex.schema,
};
